/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _PHY_DIAG_H_
#define _PHY_DIAG_H_

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include "../internal_packets/internal_packets_layouts.h"
#include "acc_reg.h"
#include "diagnostic_data.h"
#include "uphy/uphy_common.h"

//typedef vector <struct SMP_AccessRegister *>    vec_p_acceess_reg_t;
typedef vector < struct VS_DiagnosticData * >   vec_p_diagnostic_data_t;
typedef vector < vec_p_diagnostic_data_t >      vec_vec_diagnostic_data_t;
typedef list < string >                         list_string_t;

typedef map < AccRegKey *, struct VS_DiagnosticData *, bool (*)(AccRegKey *, AccRegKey *)>
        map_akey_diagnostic_data_t;
typedef vector < map_akey_diagnostic_data_t *>   vec_map_akey_diagnostic_data_t;

typedef vector <long double *>                  vec_long_double_t;

typedef vector < struct slrg_reg * >            vector_p_slrg_reg_tbl;
typedef vector < vector_p_slrg_reg_tbl >        vector_v_slrg_reg_tbl;
typedef vector < struct slrp_reg * >            vector_p_slrp_reg_tbl;
typedef vector < vector_p_slrp_reg_tbl >        vector_v_slrp_reg_tbl;
typedef vector < struct sltp_reg * >            vector_p_sltp_reg_tbl;
typedef vector < vector_p_sltp_reg_tbl >        vector_v_sltp_reg_tbl;

typedef vector < struct ptas_reg * >            vector_p_ptas_reg_tbl;

typedef vector < class AccRegHandler * > vec_pareg;
typedef vector < class DiagnosticDataInfo * > vec_diagnostic_data_t;

struct BER_thresholds_warning_error {
    double warning;
    double error;
};

typedef vector < struct BER_thresholds_warning_error > vec_ber_thresholds_val_t;
typedef map < IBFECMode, vec_ber_thresholds_val_t > map_fec_mode_ber_thresholds_val_t;

void PhyCountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
void PhyCountersResetDelegator(const clbck_data_t &clbck_data,
                               int rec_status,
                               void *p_attribute_data);
void PCICountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
void SMPAccessRegisterGetDelegator(const clbck_data_t &clbck_data,
                                   int rec_status,
                                   void *p_attribute_data);
void SMPAccessRegisterPTASGetDelegator(const clbck_data_t &clbck_data,
                                   int rec_status,
                                   void *p_attribute_data);

typedef void (*pack_port_lane_reg_t) (u_int32_t port,
                                      u_int32_t lane,
                                      u_int8_t *buff);

//data for registers with access by port and lane
typedef struct portlane_reg {
    u_int16_t            reg_id;
    u_int64_t            not_support_bit;
    pack_port_lane_reg_t pack;
    string               section;
} portlane_reg_t;

/******************************************************/
/******************************************************/
/******************************************************/
class PhyDiag : public Plugin {

    friend class Register;
    friend class SLRegister;
    friend class MPEINRegister;
    friend class MPIRRegister;
    friend class PEUCG_DLN_Register;
    friend class PEUCG_CLN_Register;

    friend class AccRegHandler;
    friend class AccRegNodeHandler;
    friend class AccRegPortHandler;
    friend class AccRegPortLaneHandler;

    friend class AccRegGroupHandler;
    friend class AccRegNodeSensorHandler;
    friend class AccRegTypeGroupHandler;
    friend class AccRegSpecificHandler;

private:
    ////////////////////
    //members
    ////////////////////
    IBFabric                     *p_discovered_fabric;
    Ibis                         *p_ibis_obj;
    IBDMExtendedInfo             *p_ibdm_extended_info;
    CapabilityModule             *p_capability_module;
    list_p_fabric_general_err     phy_errors;
    list_p_fabric_general_err     phy_effective_ber_errors;
    list_p_fabric_general_err     phy_symbol_ber_errors;

    bool                          can_send_mads_by_lid;
    bool                          to_get_phy_info;
    bool                          to_get_ppamp;
    bool                          to_dump_cap_reg;
    bool                          to_reset_counters;
    bool                          to_get_pci_info;
    bool                          to_reset_pci_counters;
    bool                          to_show_ber_warnings;
    bool                          to_load_ber_thresholds_table;
    bool                          to_get_cable_disconnected_ports;
    bool                          to_get_cable_full_data;
    int                           clbck_error_state;
    u_int32_t                     total_ber_errors;
    string                        ber_thresholds_table_file;

    vector_p_node                 nodes_vector;
    vector_p_port                 ports_vector;

    //port_cidx_and_dd_idx_to_dd_page[port_creation_idx][dd_idx]
    vec_vec_diagnostic_data_t     port_cidx_and_dd_idx_to_dd_page;
    vec_vec_diagnostic_data_t     node_cidx_and_dd_idx_to_dd_page;
    vec_map_akey_diagnostic_data_t dd_idx_and_akey_to_dd_page;

    vec_long_double_t             ber_vec;
    vec_long_double_t             eff_ber_vec;

    vec_pareg                     reg_handlers_vec;
    vec_pareg                     pci_reg_handlers_vec;
    vec_diagnostic_data_t         diagnostic_data_vec;

    map_fec_mode_ber_thresholds_val_t ber_thresholds_table;

    set_string_t                      enabled_regs;
    map_string_2_set_string_t         register_dependencies;

    list_string_t                     uphy_dumps;

    ////////////////////
    //methods
    ////////////////////
    //Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR /
    //IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
    int BuildPhyCountersDB(list_p_fabric_general_err &phy_errors,
                           u_int32_t dd_idx);
    //Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR /
    //IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
    int BuildPCICountersDB(list_p_fabric_general_err &phy_errors,
                           u_int32_t dd_idx,
                           map_akey_areg *data_map);
    /*
     * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int SMPAccRegGetByDirect(IN direct_route_t *p_direct_route,
                             IN u_int8_t port_num,
                             OUT struct SMP_AccessRegister *p_smp_acc_reg,
                             const clbck_data_t *p_clbck_data);

    int GMPAccRegGet(IN u_int16_t lid,
                     IN u_int32_t register_id,
                     OUT struct GMP_AccessRegister *p_gmp_acc_reg,
                     IN const clbck_data_t *p_clbck_data);

    inline u_int32_t getNodesVectorSize() {
        return (u_int32_t)nodes_vector.size(); }
    IBNode* getNodePtr(u_int32_t node_index);
    inline u_int32_t getPortsVectorSize() {
        return (u_int32_t)ports_vector.size(); }
    IBPort* getPortPtr(u_int32_t port_index);
    template <class OBJ_VEC_TYPE, class OBJ_TYPE>
    void addPtrToVec(OBJ_VEC_TYPE& vector_obj, OBJ_TYPE *p_obj);

    template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
    int addDataToVec(OBJ_VEC_TYPE& vector_obj,
            OBJ_TYPE *p_obj,
            DATA_VEC_TYPE& vector_data,
            DATA_TYPE& data);
    template <class OBJ_VEC_TYPE, class OBJ_TYPE>
    OBJ_TYPE * getPtrFromVec(OBJ_VEC_TYPE& vector_obj, u_int32_t idx);

    template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
    int addDataToVecInVec(OBJ_VEC_TYPE& vector_obj,
                          OBJ_TYPE *p_obj,
                          DATA_VEC_TYPE& vec_of_vectors,
                          u_int32_t data_idx,
                          DATA_TYPE& data);
    template <class OBJ_VEC_TYPE, class OBJ_TYPE>
    OBJ_TYPE * getPtrFromVecInVec(OBJ_VEC_TYPE& vec_of_vectors,
                                  u_int32_t idx1,
                                  u_int32_t idx2);

    int addDataToMapInVec(AccRegKey *p_acc_reg_key,
                          vec_map_akey_diagnostic_data_t& vec_of_maps,
                          u_int32_t data_idx,
                          struct VS_DiagnosticData& data);

    int addPhysLayerPortCounters(
    IBPort *p_port,
    struct VS_DiagnosticData &physCounters,
    u_int32_t dd_idx);

    struct VS_DiagnosticData * getPhysLayerPortCounters(u_int32_t port_index,
                                                        u_int32_t dd_idx);

    int addPhysLayerNodeCounters(IBNode *p_node,
                                 struct VS_DiagnosticData &physCounters,
                                 u_int32_t dd_idx);

    struct VS_DiagnosticData* getPhysLayerNodeCounters(u_int32_t node_index,
                                                        u_int32_t dd_idx);

    int addPhysLayerPCICounters(AccRegKey *p_acc_reg_key,
                                struct VS_DiagnosticData &physCounters,
                                u_int32_t dd_idx);

    int addBER(IBPort *p_port, long double ber);
    long double* getBER(u_int32_t port_index);

    int addEffBER(IBPort *p_port, long double ber);
    long double* getEffBER(u_int32_t port_index);

    void DumpCSVPhyCounters(CSVOut &csv_out, u_int32_t section_id);
    void DumpCSVModuleInfoCounters(CSVOut &csv_out);
    int WriteModuleInfoFile(const string &file_name);
    void DumpModuleInfo(ofstream &sout);
    void DumpCSVPCICounters(CSVOut &csv_out, u_int32_t dd_type);
    void DumpCSVRawBER(CSVOut &csv_out);
    void DumpCSVEffectiveBER(CSVOut &csv_out);
    void DumpCSVSLRGExternalInfo(CSVOut &csv_out, AccRegHandler *p_areg_handler);
    int CalcRawBER();
    int CalcEffBER(u_int64_t ber_threshold_reciprocal_val,
                   list_p_fabric_general_err& ber_errors);

    int HandleSpecialPorts(IBNode *p_curr_node,
                           IBPort *p_curr_port,
                           u_int32_t i);

    int DumpNetDumpExt();

    int DumpCSVSocketDirect();

    int DumpCSV_UPHY_Versions();
    int DumpCSV_UPHY(string section_name);

    bool isSupportFwBER(IBPort *p_port);

    void InitRegisterDependencies();
    void InitBERTable();
    const vec_ber_thresholds_val_t *GetBerThrsholdEntry(IBFECMode fec_mode) const;

    int ParseRegistersList(const string &registers_list_str);
    void CheckRegisterDependencies();
    int LoadUPHYFile(const string &filename);
    int ParseUPHYFileList(const string &uphy_path);
    int ParseUPHYDumps(const string &uphy_dumps_args);
    bool IsEnabledByFilter(const string &register_name);
    int ParseBERThresholdTable();
    int CreateBERThresholdTable(string file_path);

    int ParseAccRegPriorityValue(string value, string &result);

    void ExportToIBPort(IBPort &port, const VS_DiagnosticData *p_module_info,
                    const VS_DiagnosticData *p_latched_flag_info);

public:
    ////////////////////
    //methods
    ////////////////////
    PhyDiag(IBDiag *p_ibdiag);
    ~PhyDiag();

    //You can use this function without having PhyDiag instance.
    static inline bool IsSupportDiagnosticDataCountersFailed(u_int64_t app_data,
                                                             u_int64_t not_support_bit)
    {
        return (app_data & (NOT_SUPPORT_DIAGNOSTIC_DATA | not_support_bit));
    }


    inline void SetPhyInfoGet(bool set_val) {
        this->to_get_phy_info = set_val;
        this->MarkStage(set_val);
    }

    inline void SetPPAMPGet(bool set_val) {
        this->to_get_ppamp = set_val;
    }

    inline void SetCapRegShow(bool set_val) {
        this->to_dump_cap_reg = set_val;
    }

    inline void SetPhyInfoClear(bool set_val) {
        this->to_reset_counters = set_val;
        this->MarkStage(set_val);
    }

    inline void SetPhyCableFullData(bool set_val) {
        this->to_get_cable_full_data = set_val;
    }

    inline void SetPciInfoGet(bool set_val) {
        this->to_get_pci_info = set_val;
        this->MarkStage(set_val);
    }

    inline void SetPciInfoClear(bool set_val) {
        this->to_reset_pci_counters = set_val;
        this->MarkStage(set_val);
    }

    inline void SetShowBERWarnings(bool to_show_ber_warnings) {
        this->to_show_ber_warnings = to_show_ber_warnings;
    }

    inline void SetBERThresholdTableFile(string ber_thresholds_table_file) {
        this->to_load_ber_thresholds_table = true;
        this->ber_thresholds_table_file = ber_thresholds_table_file;
    }

    inline void SetToGetCableDisconnectedPorts(bool to_get_cable_disconnected_ports) {
        this->to_get_cable_disconnected_ports = to_get_cable_disconnected_ports;
    }

    bool IsActive();
    void Init() {}
    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit
    int HandleOption(string name, string value);

    //PhyDiag callbacks
    void PhyCountersGetClbck(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);
    void PCICountersGetClbck(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data);

    virtual bool GetFabricSummarySection(stringstream &section);
    virtual int CleanResources();

    static AccRegVia_t acc_reg_priority;
    static inline void SetAccRegPriority(AccRegVia_t val) {
        acc_reg_priority = val;
    }

    int InitPhyDataOnNodes();
};


/******************************************************/
/******************************************************/
/******************************************************/
// the class factories
extern "C" {

Plugin * create(IBDiag *p_ibdiag);
void destroy(Plugin *p);

}

#endif //_PHY_DIAG_H_
